#include "tmesh.h"
#include <fstream.h>
#include <float.h>
#include <strstrea.h>

TMesh::TMesh() {

  color = 0x00000000;
  vertsN = 0;
  verts = 0;
  trisN = 0;
  tris = 0;

}


void TMesh::Load(char *fileName) {

  ifstream ifs(fileName);
  if (ifs.ios::fail()) {
    cerr << "ERROR: cannot open: " << fileName << endl;
    return;
  }

  if (vertsN)
    delete verts;
  vertsN = 0;
  if (trisN)
    delete tris;
  trisN = 0;

  ifs >> vertsN;
  verts = new V3[vertsN];
  char lineBuffer[1000];
  ifs.getline(lineBuffer, 999);
  for (int vi = 0; vi < vertsN; vi++) {
    char lineBuffer[1000];
    ifs.getline(lineBuffer, 999);
    istrstream istr(lineBuffer);
    istr >> verts[vi];
  }
  cerr << "INFO: read " << vertsN << " vertices" << endl;


  ifs >> trisN;
  tris = new unsigned int[trisN*3];
  ifs.getline(lineBuffer, 999);
  for (int ti = 0; ti < trisN; ti++) {
    char lineBuffer[1000];
    ifs.getline(lineBuffer, 999);
    istrstream istr(lineBuffer);
    istr >> tris[3*ti+0] >> tris[3*ti+1] >> tris[3*ti+2];
  }

  cerr << "INFO: read " << trisN << " triangles" << endl;

  ifs.close();

}


void TMesh::RenderFM(FrameBuffer *fb, PHC *phc) {

  V3 *projVerts = new V3[vertsN];
  for (int vi = 0; vi < vertsN; vi++) {
    phc->Project(verts[vi], projVerts[vi]);
  }

  for (int tri = 0; tri < trisN; tri++) {

    // projection
    V3 projVs[3];
    bool triangleOK = true;
    for (int vi = 0; vi < 3; vi++) {
      projVs[vi] = projVerts[tris[3*tri+vi]];
      if (projVs[vi][0] == FLT_MAX) {
        triangleOK = false;
        break;
      }
    }
    if (!triangleOK)
      continue;

    // find axes aligned bounding box
    BBox bbox(projVs[0]);
    bbox.AddPoint(projVs[1]);
    bbox.AddPoint(projVs[2]);
    
    int left = (int) (bbox.corners[0][0] + 0.5f);
    int right = (int) (bbox.corners[1][0] - 0.5f);
    int top = (int) (bbox.corners[0][1] + 0.5f);
    int bottom = (int) (bbox.corners[1][1] - 0.5f);

    if (left < 0)
      left = 0;
    if (right < 0)
      continue;
    if (top < 0)
      top = 0;
    if (bottom < 0)
      continue;
    if (left > fb->w)
      continue;
    if (right > fb->w)
      right = fb->w;
    if (bottom > fb->h)
      bottom = fb->h;
    if (top > fb->h)
      continue;


    // rasterization setup
    V3 eeqs[3];
    for (int ei = 0; ei < 3; ei++) {
      int _ei = (ei+1)%3;
      eeqs[ei][0] = projVs[_ei][1]-projVs[ei][1];
      eeqs[ei][1] = projVs[ei][0]-projVs[_ei][0];
      eeqs[ei][2] = -projVs[ei][1]*eeqs[ei][1] - projVs[ei][0]*eeqs[ei][0];
      int __ei = (_ei+1)%3;
      V3 P2 = projVs[__ei];
      P2[2] = 1.0f;
      if (eeqs[ei]*P2 < 0.0f) {
        eeqs[ei] = eeqs[ei]*-1.0f;
      }
    }

    // rasterization loops
    for (int v = top; v <= bottom; v++) {
      for (int u = left; u <= right; u++) {
        V3 pixPoint(0.5f+(float)u, 0.5f+(float)v, 1.0f);
        if (
          eeqs[0]*pixPoint < 0 ||
          eeqs[1]*pixPoint < 0 ||
          eeqs[2]*pixPoint < 0)
          continue;

        // we are inside triangle

        // we are visible

        fb->Set(u, v, color);

      }
    }

  }

  delete projVerts;
  

}

void TMesh::RenderWF(FrameBuffer *fb, PHC *phc) {

  for (int tri = 0; tri < trisN; tri++) {
    for (int vi = 0; vi < 3; vi++) {
      fb->DrawSegment3D(verts[tris[3*tri+vi]], verts[tris[3*tri+((vi+1)%3)]], 
        *phc, color);
    }
  }

}


V3 TMesh::FindCenter() {

  V3 ret(0.0f, 0.0f, 0.0f);
  for (int i = 0; i < vertsN; i++) {
    ret = ret + verts[i];
  }
  return ret / (float) vertsN;

}


void TMesh::ScaleToMaxDim(float maxDim) {

  int vi;
  BBox bbox(verts[0]);
  for (vi = 0; vi < vertsN; vi++) {
    bbox.AddPoint(verts[vi]);
  }

  float oldMaxDim = bbox.GetDiagonal();

  float scaleFactor = maxDim/oldMaxDim;

  V3 center = FindCenter();
  TranslateBy(center);
  ScaleBy(scaleFactor);
  TranslateBy(center*-1.0f);
  
}


void TMesh::TranslateBy(V3 translationVector) {

  for (int vi = 0; vi < vertsN; vi++) {
    verts[vi] = verts[vi] + translationVector;
  }

}

void TMesh::TranslateTo(V3 newCenter) {

  V3 center = FindCenter();
  TranslateBy(newCenter-center);

}

void TMesh::ScaleBy(float scf) {

  for (int vi = 0; vi < vertsN; vi++) {
    verts[vi] = verts[vi] * scf;
  }

}
